##############################################################################
# TAG BIBLE VERSES IN HTML FILES
##############################################################################
#
# (c) 2003 www.SpiritAndTruth.org
# May be freely used without restriction.
# mailto:contact@SpiritAndTruth.org
#
#
#############################################################################
#
# See the HELP string below for a description of the various flags.
#
# This program is a general-purpose tagging utility which scans
# files for recognizable bible addresses and then marks those bible
# addresses a user-defined way.  It is typically used to descend a
# directory containing HTML files and to tag addresses with URLs
# which allow automatic verse lookup. But the program could be used
# to tag addresses in just about any sort of file with more
# general markup.  See the '-tag' and -tagvars flags below which allows you to
# describe how addresses are to be tagged.
#
# One caveat concerning tagging files: if you use a relative URL to tag a file
# to perform scripture lookup and that file gets moved elsewhere, 
# the URL will typically break.  We 'solve' this problem by inserting
# a javascript file in each tagged file. This javascript file defines
# a function called "X()" which is called to perform URL lookup.
#
# The function tests for the existence of a 'lookup.js" file in the
# urlPath directory. If the 'lookup.js' file exists, then
# the X() function uses the relative URL since the tagged file appears
# to be in its original location.
#
# If the 'lookup.js' file is missing, then we assume that the tagged file
# has been moved (probaby downloaded from the website by a user
# and we convert any relative URL into an absolute URL pointing
# at the lookup at our website.
#
# This approach has the following advantages:
#
#    1. URLs in tagged files continue to use relative paths
#       when the files are in situ on the website or on a CDROM.
#
#    2. URLs in tagged files which have been moved elsewhere
#       automatically redirect to www.SpiritAndTruth.org/lookup/index.htm
#
#    3. No conversion processes are necessary when building directories.
#
# Alternative approaches have drawbacks: relative URLs will break
# when users download the file from the website; absolute URLs
# won't run from the stand-alone CD without an internet connection.

HELP = '''
Usage: bibletag [flags] path1 [path2...]

TYPICAL USAGE
-------------

The most common way which bibletag is used is to tag HTML 
files containing Bible references for one of two 
destinations:

-st         Convert Bible addresses into links to the
            dynamic lookup Bible on the internet at
            www.SpiritAndTruth.org/lookup/index.htm.

-lbxst      A combination of -lbx and -st tagging
            where the address acts as a link to the
            www.SpiritAndTruth.org bible and an icon 
            links to Libronix on the local machine.

-gospelcom  Tag Bible addresses with Bible links to
            the bible.gospelcom.net website.

-lbxgospel  A combination of -lbx and -gospelcom tagging
            where the address acts as a link to
            bible.gospelcom.net and an icon links to
            Libronix on the local machine.

-studylight Tag Bible addresses with Bible links to
            the www.studylight.org website.

-lbxstudy   A combination of -lbx and -studylight tagging
            where the address acts as a link to
            www.studylight.org and an icon links to
            Libronix on the local machine.

-lbx        Tag Bible addresses to launch Libronix on
            the local machine.

-lbxpbb     Tag Bible addresses with Bible links suitable
            for subsequent compilation by the Libronix
            Personal Book Builder.

For example, to tag a file in preparation for compilation
by the Libronix Personal Book Builder:

  bibletag -lbxpbb 'c:\my directory\myfile.htm'

(Quotes are required around paths containing
spaces.)

RUNNING BIBLETAG
----------------

Bibletag comes in two flavors: a python source file (bibletag.py)
and a DOS/Windows program (bibletag.exe).

If you are unfamiliar with Python, then you probably want to use
the DOS/Windows program. However, bibletag does not run with a
graphical user interface (e.g., window).

It is possible to convert files by dragging and dropping them onto the
bibletag.exe program as follows:

1. Use Windows Explorer (the file manager) to open the folder containing
   bibletag.exe. We'll call this FOLDER1.

2. Open the flags.txt document and edit the flags which you want
   to pass to bibletag when it tags your file(s). (See the other
   flags_*.txt files for examples.)

   NOTE: If you are running bibletag using python or from a DOS command
   window and are specifying all your flags on the command line,
   then you will want to delete the flags.txt file. Otherwise, the flags
   from flags.txt will be combined with the flags specified on 
   the command line.

3. To convert a file (or folder containing a group of files): 

    A. Open another window in Windows Explorer.
    B. Move to the folder containing the file or subfolder you want to
       tag. We'll call this FOLDER2.
    C. Drag the file(s) or subfolder from within FOLDER2 that you want
       to tag.
    D. Drop them onto the bibletag.exe file in FOLDER1. Be sure you drop
       them right on the bibletag.exe file.

4. Bibletag will run and convert your files, leaving the output in the
   same location as the original file(s) or folder(s), but with an additional
   __bibletag extension to the file name. For example, if the input file
   was:

        c:\My Documents\luke.htm

   bibletag will produce:

        c:\My Documents\luke_bibletag.htm

5. When bibletag runs, it logs its activity to the file 'bibletag_log.txt'
   in the same folder bibletag.exe resides.

FLAGS
-----

flags.txt       When a 'flags.txt' file exists in the bibletag folder,
                then any flags specified on the command line are combined
                with flags listed in this file.

                The flags in the flags.txt file may be all on one line:

                    -book Rev -exact -yes

                or individually on separate lines:

                    -book Rev
                    -exact
                    -yes

                The flags.txt file provides a way for common flags to be listed
                in one place and, more importantly, provides a way to pass
                flags to the Windows/DOS command-line version of the program
                when running it using drag-and-drop (see above).

-backup [dir]   Directory where original files are optionally saved
                prior to tagging Bible addresses.
                -------------------------------------------------------
                THIS FLAG IS BEING RETAINED FOR BACKWARD COMPATIBILITY. 
                -------------------------------------------------------
                We recommend discontinuing this flag in favor of the new
                default behavior which does not modify the original files
                but creates new files for the tagged output.

                If no -backup flag is supplied, the tagging operation 
                leaves the input files unmodified and produces 
                correspondingly-named output files. For example,
                if the -backup flag is omitted, then an input file with
                the name:

                    c:\My Documents\myFile.htm

                will be tagged to produce the output file:

                    c:\My Documents\myFile_bibletag.htm

-book name      Assumes addresses which appear without an
                associated Bible book are in the specified book. 
                For example '-book Re' will treat all floating
                addresses as from the Book of Revelation. Numbers
                are considered floating addresses when they contain
                a colon separator (e.g., 12:45) or are preceded by
                a chapter or verse "introducer" (e.g., "chapter"
                "chapters", "chs." "verse", "verses" "vv." ).

-books file     Optional file containing addition bible book names
                to be recognized. The file must contain 66 lines of
                text with one book name per line corresponding, in order,
                to each of the 66 books of the bible in their standard
                arrangment.  See the example file 'books.txt'.

-exact          Only tag addresses which exactly match bible book
                names (including those specified in the optional
                -books file). See the -ignorecase flag.

-ext            Specify the extension for files to be tagged.
                The default extension is '.htm?'
                The extension is case-insensitive.

-gospelcom      Links bible references to the bible.gospelcom.net website.

-help  -?       Display this help information and exit.

-history        Display revision history and exit.

-ignorecase     Ignore the alphabetic case of Bible book names
                (including those specified in the optional
                -books file) when tagging addresses. See the -exact flag.

-insert file    Inserts the contents of the specified file
                prior to the '<body>' tag (if any). This is
                useful for inserting javascript into each file
                which the tag can then utilize.

                This has largely been superceded by the -tagjs
                flag.

                The javascript code is inserted after any code 
                inserted by the -tagjs flag so you can redefined
                the X() function to customize operation.

-join           Preserve joined lines in output.
                Input lines are joined as paragraphs during 
                tagging in order to successfully recognize 
                addresses which span line boundaries. After
                tagging, lines are separated to preserve 
                their original format. Use this flag to retain
                joined lines in the output. (May be required for
                non-HTML files where tags cannot span line breaks.)
                See -nojoin.

-lbx            Tag Bible addresses to launch Libronix on
                the local machine.
                LibronixLink.gif file in the same directory as
                the tagged file before use.)

-lbxgospel      A combination of -lbx and -gospelcom tagging
                where the address acts as a link to
                bible.gospelcom.net and an icon links to
                Libronix on the local machine.  (Be sure to place the
                LibronixLink.gif file in the same directory as
                the tagged file before use.)

-lbxgospel+     Same as the -lbxgospel flag, but substitutes
                a "+" sign in the place of the icon for
                each link. This is less visually distracting 
                in some applications.

-lbxstudy       A combination of -lbx and -studylight tagging
                where the address acts as a link to
                www.studylight.org and an icon links to
                Libronix on the local machine.  (Be sure to place the
                LibronixLink.gif file in the same directory as
                the tagged file before use.)

-lbxstudy+      Same as the -lbxstudy flag, but substitutes
                a "+" sign in the place of the icon for
                each link. This is less visually distracting 
                in some applications.

-lbxpbb         Tags address fields in preparation for use in the
                www.logos.com Libronix Personal Book Builder compiler.
                For example, text of the form 
                
                Genesis 1:1, 2; 3:15

                will be tagged as

                Genesis [[1.1|bible.1.1.1]], [[2|bible.1.1.2]]; [[3:15|bible.1.3.15]]

                When the -lbxpbb flag is specified, Strong's Greek and Hebrew
                numbers are also tagged. For example:

                    h123 --> [[HebrewStrongs:h123]]
                    G439 --> [[GreekStrongs:G439]]

-lbxst          A combination of -lbx and -st tagging
                where the address acts as a link to
                bible.gospelcom.net and an icon links to
                Libronix on the local machine. (Be sure to place the
                LibronixLink.gif file in the same directory as
                the tagged file before use.)

-lbxst+         Same as the -lbxst flag, but substitutes
                a "+" sign in place of the icon for each link.
                This is less visually distracting in some
                applications.

-lbxc filename  Generates an empty HTML commentary for every verse
                of the bible tagged with milestones for the Libronix
                Personal Book Builder compiler. Includes a "styles.css"
                file to define appearance of content. The filename
                specifies the name of the file to be built.
                See the "commentary.htm" file provided as an example.

-lbxcnov filename Same as -lbxc, but without verse milestones.

-nojoin         Do not join lines.
                Prevents the tagger from joining lines into one
                single large line before scanning for bible addresses.
                The input file cannot contain addresses which span
                across line boundaries.
                See -join.

-olivetree      Tags address fields in preparation for use in the
                www.olivetree.com Biblereader program.

path1 path2...  A list of directories or files which are searched
                for files to be tagged.

-plainlinks     When operating on an HTML file, modify the style
                in the document header so that URLs appear without
                an underline except when they are hovered over with
                the mouse.

-range          Addresses which specify a chapter without a
                designated verse use a tag which includes the
                entire verse range for that chapter rather than
                verse 1.  For example, an address such as "Rev. 1"
                which would normally result in a tag of "Rev. 1:1"
                would result in a tag of "Rev. 1:1-20" instead.

-scrub          Scrubs out artifacts generated by Microsoft Word 
                DOC-to-HTML conversion. The following conversions
                are applied which also dramatically reduce file size:
                    mdash (becomes dash)
                    ndash (becomes dash)
                    <st1:*>..</st1:*> (smart tags deleted)
                    @list lN (list definitions deleted)
                    span.BAS* (style definitions deleted)

-st             Enables flags typically used by SpiritAndTruth.org.
                Equivalent to: '-scrub -tagjs -plainlinks'

-swordsearcher  Tags address fields in preparation for use in the
                www.swordsearcher.com SwordSearcher program.
                Automatically enables the -tagbook flag.

-tag '1|2|3'    Optional custom tag string to be used to surround 
                Bible addresses. Unless specified otherwise, the
                tag string used is as follows:

                <a href="https://www.spiritandtruth.org/lookup/index.htm?|">|</a>

                This will tag the address string 'John 3:16' as follows:

                'John <a href="https://www.spiritandtruth.org/lookup/index.htm?John 3:16">3:16</a>'

                This will convert Bible references into dynamic links to the 
                related verse at the SpiritAndTruth.org website.

                For default operation, this is as if the following values 
                had been specified.

                1 = '<a href="https://www.spiritandtruth.org/lookup/?'
                2 = '">'
                3 = '</a>'

                Tags which contain URLs which are relative (no ':') are
                automatically adjusted for the location of the file
                being scanned within its directory tree.  For example,
                specifying a tag with 'href=file.txt' will be converted
                to 'href=../file.txt', 'hef=../../file.txt' and so on
                files are tagged which are progressively deeper in the
                directory tree.

-tagvar string  An variable form of the tag string may be specified
                which provides the ability to define the tag using the 
                following variables:

                    %%BOOK%% - the abbreviated bible book name
                    %%BOOKLONG%% - the full canonical bible book name.
                    %%CHAPTER%% - the chapter.
                    %%VERSE%% - the verse.
                    %%TOPIC%% - the topic being tagged (see -topics flag)
                    %%BIBLEADDRESS%% = "%%BOOKLONG%% %%CHAPTER%%:%%VERSE%%"
                    %%MATCH%% - the text matched in the original file

                Each of the variables may be specified in a quoted form
                suitable for embedding in a URL:

                    %%QUOTEDBOOK%%, %%QUOTEDBOOKLONG%, %%QUOTEDTOPIC%%, etc.

                For example, a tag could be defined as follows:

                -tagvar <a href="https://www.spiritandtruth.org/lookup/index.htm?%%QUOTEDBIBLEADDRESS%%">%%MATCH%%</a>

                Tags which contain URLs which are relative (no ':') are
                automatically adjusted for the location of the file
                being scanned within its directory tree.  For example,
                specifying a tag with 'href=file.txt' will be converted
                to 'href=../file.txt', 'hef=../../file.txt' and so on
                files are tagged which are progressively deeper in the
                directory tree.

-tagjs [urlPath] Use javascript to implement the tagging and lookup.
                 Tags within the document use an abbreviated query form,
                 as if you had specified 
                 
                 -tag '<a href=javascript:X("|")>|</a>'

                 Javascript code is inserted prior to the <body> of the 
                 document which defines function "X()" to prepend the 
                 specified urlPath to each query before performing the 
                 lookup. With this scheme, a tag for 'John 3:16' would 
                 have the following javascript URL:

                 href=javascript:X("John 3:16")

                 The actual lookup would go to: 'urlPath?John 3:16'

                 If urlPath is omitted, it defaults to 
                 'https://www.SpiritAndTruth.org/lookup/index.htm'.

                 The javascript code is inserted ahead of any file
                 specified by the -insert flag so you can redefined
                 the X() function to customize operation. The
                 javascript variable 'urlPath' contains the url Path.

                 If the urlPath is relative, it is adjusted to reflect
                 the relative position of each file within the tagged directory 
                 structure.

                 If the document is displayed within a containing frame which
                 contains a pane named 'bible', then the tagged address is
                 looked up within that pane. This allows the document and bible
                 reference to be viewed simultaneously.

                 If the urlPath is relative and points to a directory which
                 is not found, the assumption is made that the file has
                 been moved from its original location and the lookup
                 occurs at https://www.SpiritAndTruth.org/lookup/index.htm
                 instead.

                 This allows tagged files to reside within a directory
                 structure which includes support for bible verse lookup
                 locally. But if the files get moved (e.g., downloaded
                 from your server), the bible lookup links which were
                 previously relative, will not operate as if absolute
                 so the lookups will not break, but will go over the
                 internet instead.

-tagbook        The book name will be encapsulated in the tag (normally
                only the address portion is tagged). Automatically
                enabled when -swordsearcher is specified.

-topics file    A file containing a list of topics to be tagged.

                When specifying this flag, either a custom "-tagvar string" 
                must be provided (which contains a %%TOPIC%% variable describing
                how each topic should be tagged in the output). Alternately,
                one of the following tag flags must be specified: 
                -lbx, -lbxgospel, -lbxgospel+, -lbxpbb, -lbxst, -lbxst+, -st,
                -lbxstudy, -lbxstudy+.

                (When an -lbx* flag is specified the link to Libronix will treat
                the topic as a "Text" data type so you will want to create
                a keylink to your favorite dictionary using that data type.)

                The topics file contains a list of topics, one per line.
                Each topic line contains one or more phrases separated
                by the vertical bar character (|). For example:

                    Temple|Temple Mount|House of God|God\'s House|...

                Only the first phrase is required and specifies the 
                string which will be used to expand the %%TOPIC%% variable. 
                All the phrases match when scanning for a topic in the 
                text. The result is that a variety of phrases are all 
                tagged with the primary (first) phrase in the output.
                Phrases are case-insensitive.

                Special wildcard characters may be used:

                    ? = matches any character
                    * = matches any sequence of characters
                    + = matches one or more of the previous character

                So for instance, to match "curse" "curseth" "cursed"
                and other similar variations, specify:

                    curs*

-v              Verbose. List each address as it is tagged.

-ver            Display revision history and exit.

-yes            Don't ask for confirmation before tagging each file.

Developed by: www.SpiritAndTruth.org contact@SpiritAndTruth.org
'''

VERSION = '1.29'
HISTORY = '''

1.29 Added -lbxcnov flag to generate empty commentary without verse milestones.

1.28 Fixed -plainlinks flag to insert link style after any styles
     existing in the document to be sure they take precendence.

1.27 Bible addresses which span a range of verses (e.g., Gen. 1:3-7)
     now include the entire range in the tagged information. (Previous
     versions displayed the range, but represented the range by only
     including the first verse in the tag.)

1.26 Added the following new flags:

     -ignorecase 
      Ignore alphabetic case when searching for Bible book names.

     -studylight 
      Tag addresses to point to www.StudyLight.org Bible lookup.

     -lbxstudy
      Combine the -studylight flag with the -lbx flag (which see).

     -lbxstudy
      Combine the -studylight flag with the -lbx flag (which see)
      using a "+" sign in place of the Libronix icon.

      Fixed several defects.

1.25 1. Added support for specifying flags in a 'flags.txt' file so that
        bibletag can be run by dragging and dropping files using Windows 
        Explorer.

        Modified backing up files so that the default operation now
        leaves the input files unmodified and produces correspondingly-named
        output files. If the -backup flag is omitted, then an input file with
        the name:

           c:\My Documents\myFile.htm

        will be tagged to produce the output file:

           c:\My Documents\myFile_bibletag.htm

     2. The -st flag no longer activates the -backup flag.

     3. Bibletag now creates a 'bibletag_log.txt' file in the folder it 
        resides in where it keeps a log of its activities. This makes it
        possible to view the results of drag-and-drop operations.

1.24 When scanning HTML or XML files, we specifically avoid scanning
     or joining all lines prior to the <body> marker.

1.23 Added the -lbxc flag which generates an empty Libronix
     Personal Commentary. See documentation for -lbxc flag.
     
1.22 Fixed a problem where the '-tagbook' flag was not working

1.21 Added -tagvar and -topics flags. The -tagvar flag allows
     a custom tag to be defined which can include variables
     which refer to the matched book, chapter, verse, topic, etc.
     It provides maximum flexibility in marking up a text.
     The -topics flag allows the specification of topic strings
     for tagging.  See the description of each flag for further
     information.

1.20 We now provide warnings when ignoring out-of-range
     chapter or verse numbers. Added -book flag (which see).

1.19 Added -lbxst+ and -lbxgospel+ flags (which see).

1.18 When the Libronix (-lbxpbb) flag is specified, Strong's
     number of the form H1234, h1234, G1234, g1234 are
     tagged with Libronix Strong's number links.

1.17 Fixed some minor tagging issues. Previous tags are now
     ignored by -lbxpbb flag so retagging of a previously
     tagged file is safe.

1.16 Added -lbxpbb flag to automatically generate bible tags in
     Libronix bible linke format (e.g., "Genesis 1:1" appears as
     "Genesis [[1:1|bible.1.1.1]]"). See www.logos.com

1.15 Added -plainlinks flag to support non-underlined URLs in HTML
     output. Links are only underlined on hover over.

1.14 Fixed address parsing in address.py to allow verse numbers to
     include 'a' 'b' 'c' or 'd' subscript.

1.13 Added -olivetree and -swordsearcher flags.

1.12 Fixed defect which allowed verses beyond the last valid verse
     number of a chapter to be misidentified as bible addresses.

1.11 Added ability to detect 'vs.' as a verse indicator.
     Now match book names which have an undetermined amount of
     whitespace between book number and name (e.g., "1    Corinthians").

1.9  Fixed error in parsing verses within a single chapter separated
     by a comma (but without whitespace).

1.8  Added 'range' flag. Modified address format within tag to
     use plain format (e.g., "John 3:16") rather than semi-custom 
     encoding (e.g., "John_C3V16").

1.7  Fixed javascript error in code generated by 'tagjs' flag.

1.6  The code inserted for 'tagjs' now includes a URL to disable the
     separate bible column, if present.

1.4, 1.5 Fixed more minor address recognition problems where certain
         textual conditions caused addresses to be incorrectly recognized.

1.3   Added -tagjs flag for more efficient java-based tagging.

1.2   Revised address pattern matching code to recognize a wider
      variety of book number designators (1st, 2nd, 3rd, I, II, III,
      first, second, third) and to assume floating addresses following
      the words 'v.', 'vv.', 'verse[s]' are verses rather than
      chapter numbers.

1.1   When exact matching is operating, books prefixed with 'first',
      'second', 'third', and 'I' through 'III' are not matched unless
      a -book <file> is specified which contains those explicit
      strings.

      Added -history and -ver flags.

1.0   Initial version.
'''

import sys
import time
import string
import re
import os.path
import shutil
import urllib
import exceptions
import address

##############################################################################
# Visit a directory when walking a directory tree.
##############################################################################
def visit( processor, dirname, names ) :
    # Visit each directory below a directory tree.

    #print 'Visiting directory:', dirname
    for name in names :
        processor.scanFile( os.path.join( dirname, name ) )

    return


##############################################################################
# Identify and tag bible addresses within files.
##############################################################################
class Processor :

    def __init__( self, args ) :
        # Perform tagging.

        self.m_ask = 1
        self.m_backupDir = ''
        self.m_booksFileName = ''
        self.m_books = None
        self.m_book = ''
        self.m_exactMatch = address.AddressFinder.MATCH_DEFAULT
        self.m_olivetree = 0
        self.m_lbx = 0
        self.m_lbxpbb = 0
        self.m_lbxc = ''
        self.m_lbxcnov = ''
        self.m_lbxst = 0
        self.m_lbxstplus = 0
        self.m_lbxgospel = 0
        self.m_lbxgospelplus = 0
        self.m_lbxstudy = 0
        self.m_lbxstudyplus = 0
        self.m_gospelcom = 0
        self.m_studylight = 0
        self.m_swordsearcher = 0
        self.m_verbose = 0
        self.m_st = 0
        self.m_insertFileName = ''
        self.m_insertFileContents = ''
        self.m_join = 0
        self.m_nojoin = 0
        self.m_scrub = 0
        self.m_scrubs = []
        self.m_range = 0
        self.m_walkFromPath = ''

        self.m_ext = '.htm(l)?'
        self.m_paths = []
        self.m_defaultLookup = 'https://www.spiritandtruth.org/lookup/index.htm'
        self.m_tag = [ '<a title="View passage at www.SpiritAndTruth.org" href="' + self.m_defaultLookup + '?', '">', '</a>' ]
        self.m_tagvar = ''
        self.m_tagvaraddr = ''
        self.m_topics = []
        self.m_topicMarker = 'XYZZYX'
        self.m_tagbook = 0
        self.m_plainlinks = ''
        self.m_mndash = '([\x96\x97\xbe]|&[mn]dash;)'
        self.m_insertExp = re.compile( '<body', re.IGNORECASE )

        # Javascript-related tag information.
        self.m_tagjsPath = ''
        self.m_tagjsCode = '''
<script language="javascript" src="%URLDIR%/lookup.js"> </script>
<script language="javascript">
X = function( find )
    {
    var url = "%URLPATH%"
    if ( ! window.lookup )
        {
        if ( -1 == url.indexOf( ':' ) )
            {
            // url = 'https://www.spiritandtruth.org/lookup/index.htm'
            // Version without split window.
            url = 'https://www.spiritandtruth.org/lookup/index-.htm'
            }
        }
    url = url + '?' + find
    var pane = window
    if ( parent.bible )
        {
        pane = parent.bible
        }
    pane.location.href = url
    }
if ( parent.bible )
    {
    var myDoc = window.location.href
    window.document.writeln( '<center><a href="' + myDoc + '" target="_parent">VIEW ARTICLE WITHOUT BIBLE</a><HR></center>') 
    }
</script>
'''
        # patterns for converting relative URLs.
        urlIntro = '(href=[\"\']?(javascript[^(]+\([\"\'])?)'
        self.m_urlIntroExp = re.compile( urlIntro, re.IGNORECASE )
        self.m_relUrlExp = re.compile( urlIntro + '(?=[^:>]+$)', re.IGNORECASE )

        #self.m_addrExp = r'(?P<book>([1-3]\s?)?[A-Za-z\.]+)\s+(?P<chapter>\d+):(?P<verse>\d+)'
        #self.m_addrPat = re.compile( addrExp )

        # What directory are we running from?
        self.m_hereDir = os.path.dirname( args[ 0 ] )

        # Build log file
        self.m_logFile = None
        self.m_logPath = self.path( 'bibletag_log.txt' )
        self.m_logFile = open( self.m_logPath, 'w' )
        self.log( 'BIBLETAG LOG - contact@SpiritAndTruth.org' )
        self.log( time.ctime() )
        self.log( 'ARGS: %s' % args )

        # Decode arguments on command line.
        argsWithoutFlags = self.decodeArgs( args[ 1 : ] )

        # Decode arguments in any associated 'flags.txt' file.
        flagFile = None
        try :
            flagPath = self.path( 'flags.txt' )
            if os.path.exists( flagPath ) :
                flagFile = open( flagPath, 'r' )
                self.log( 'Reading flags from file: %s' % flagPath )
                flags = []
                for line in flagFile.readlines() :
                    self.log( '\t%s' % line )

                    # protect quoted strings during split.
                    output = ''
                    input = line
                    while 1 :
                        output = re.sub( '"([^"]+)\s+([^"]+)"', r'"\1_SPACE_\2"', input )
                        output = re.sub( "'([^']+)\s+([^']+)'", r"'\1_SPACE_\2'", output )
                        if output == input :  break
                        input = output

                    words = string.split( output )
                    for word in words :

                        # Remove outermost quotes, if any.
                        word = re.sub( "^'(.+)'$", r'\1', word )
                        word = re.sub( '^"(.+)"$', r'\1', word )

                        word = re.sub( '_SPACE_', ' ', word )

                        flags.append( word )

                argsWithoutFlags.extend( self.decodeArgs( flags ) )
        finally:
            if flagFile: flagFile.close()

        # Everything else must be a path.
        self.m_paths = argsWithoutFlags
        if not self.m_paths and not self.m_lbxc and not self.m_lbxcnov :
            print 'No file paths specified!'
            self.printHelp()
            self.exit( 1 )

        # Ask to continue if necessary.
        if ( self.m_ask ) :
            answer = raw_input( 'Continue? [y] ' )
            if answer and 'y' != answer[ 0 ] and 'Y' != answer[ 0 ] : 
                self.log( 'Aborting!' )
                self.exit( 1 )

        # Are we generating an empty commentary?
        commentary = self.m_lbxc 
        if not commentary : commentary = self.m_lbxcnov
        if commentary :
            # Open output file.
            commentaryPath = self.path( commentary )
            file = open( commentaryPath, 'w' )
            file.write( '<html><head><title>\n')
            file.write( 'Libronix Personal Commentary\n' )
            file.write( '</title>\n')
            file.write( '<link rel=stylesheet type="text/css" HREF="styles.css">\n' )
            file.write( '</head><body>\n')
            file.write( '<h1>Libronix Personal Commentary</h1>\n' )
            iBook = 0
            for bookName in address.Address.BOOKNAMES :
                print address.Address.BOOKNAMESLONG[ iBook ]
                file.write( '<h2>%s</h2>\n' % address.Address.BOOKNAMESLONG[ iBook ] )
                maxChapter = address.Address.BOOKS[ bookName ]
                iMaxChapter = int( maxChapter )
                for iChapter in range( 1, iMaxChapter + 1 ) :
                    chapter = str( iChapter )
                    file.write( '<hr><h3>%s %s</h3>\n' % ( address.Address.BOOKNAMESLONG[ iBook ], chapter ) )
                    maxVerse = address.Address.RANGES[ bookName + chapter ]
                    iMaxVerse = int( maxVerse )
                    # Verse content is optional.
                    for iVerse in range( 1, maxVerse + 1 ) :
                        verse = str( iVerse )
                        addr = address.Address( bookName, chapter, verse )
                        file.write( addr.lbxMilestone() + '\n' )
                        file.write( '<h4>%s</h4>\n' % addr.short() )
                        # Only write first verse anchor if skipping verses.
                        if self.m_lbxcnov : break
                        file.write( '<p>(No text at this verse.)</p>\n' )
                iBook += 1
            file.write( '</body></html>\n' )
            file.close()
            self.exit( 0 )

        # Are we scrubbing the text?
        if self.m_scrub : 
            # Convert m and n dash to normal dash.
            self.m_scrubs.append( [ re.compile( self.m_mndash, re.IGNORECASE ), '-' ] )
            self.m_scrubs.append( [ re.compile( '</?st1:[^>]*>', re.IGNORECASE ), '' ] )
            self.m_scrubs.append( [ re.compile( '@list [^}]+}', re.IGNORECASE ), '' ] )
            self.m_scrubs.append( [ re.compile( 'span.BAS[^}]+}', re.IGNORECASE ), '' ] )

            # Private styles we don't typically use.
            if 0 :
                self.m_scrubs.append( [ re.compile( 'p.Answer[^}]+}', re.IGNORECASE ), '' ] )
                self.m_scrubs.append( [ re.compile( 'p.Question[^}]+}', re.IGNORECASE ), '' ] )
                self.m_scrubs.append( [ re.compile( 'p.NormalDouble[^}]+}', re.IGNORECASE ), '' ] )
                self.m_scrubs.append( [ re.compile( 'p.NormalLatin[^}]+}', re.IGNORECASE ), '' ] )
                self.m_scrubs.append( [ re.compile( 'p.TableHeadingSmall[^}]+}', re.IGNORECASE ), '' ] )
                self.m_scrubs.append( [ re.compile( 'p.TableEntrySmall[^}]+}', re.IGNORECASE ), '' ] )
                self.m_scrubs.append( [ re.compile( 'p.[1-9]-Double[^}]+}', re.IGNORECASE ), '' ] )
                self.m_scrubs.append( [ re.compile( 'p.[1-9]-Normal[^}]+}', re.IGNORECASE ), '' ] )

        # Filter out all HTML except anchors (we need to avoid
        # modifying text between anchors).
        # Don't modify header or title (including the body of either).
        header = '<(header|title).*</(title|header)[^>]*>'
        notAnchor = '(<(?!(/)?[Aa])[^>]*>)'
        entity= '(&\S+;)' 
        white = '(\s)'
        # Header must appear first so is found prior to notAnchor which
        # erases the tags that header must match
        htmlPat = header + '|' + notAnchor + '|' + entity + '|' + white

        # Pattern to recognize Libronix milestones and links.
        lbxPat = r'\[\[[^\]]+\]\]'
        self.m_lbxpbbExp = re.compile( lbxPat, re.IGNORECASE )

        # Libronix Strong's number pattern.
        lbxpbbStrongsPat  = r'\b(?P<prefix>[gh])(?P<number>\d+)\b'
        self.m_lbxpbbStrongsExp = re.compile( lbxpbbStrongsPat, re.IGNORECASE )

        # If lines are being joined.
        if not self.m_nojoin :
            hangingOpen = '(<(?!(/)?[Aa])[^>]*$)'
            hangingClose = '([^<>]+(?=>))'
            htmlPat = htmlPat + '|' + hangingOpen
            htmlPat = htmlPat + '|' + hangingClose
        self.m_htmlExp = re.compile( htmlPat, re.IGNORECASE )


        # Get ready to scan for addresses.
        verbose = self.m_verbose
        if self.m_verbose and self.m_logFile : 
            verbose = self.m_logFile
        self.m_addressFinder = address.AddressFinder( verbose, self.m_exactMatch, self.m_books, self.m_book )

        # For each name to be scanned for tags.
        for name in self.m_paths :

            # Get absolute path relative to starting directory.
            path = self.path( name )

            # Walk directories.
            self.m_walkFromPath = self.path()
            if os.path.isdir( path ) : 
                self.m_walkFromPath = path
                os.path.walk( path, visit, self )

            # Scan individual files.
            elif os.path.isfile( path ) : 
                self.scanFile( path )

            # Not a directory or file so warn.
            else :
                self.log( 'Ignoring Nonexistent path: %s' % path )
                self.exit( 1 )

    def log( self, text ) :
        if self.m_logFile : self.m_logFile.write( text + '\n' )
        print text

    def path( self, path = '' ) :
        # Attempt to find the file as an absolute path.

        # If a pathname was specified.
        if path :

            # If it doesn't exist from this location.
            if not os.path.exists( path ) :


                # Assume it is relative to our current directory.
                path = os.path.join( self.m_hereDir, path )

        # Convert to absolute path.
        path = os.path.abspath( path )

        # Normalize
        path = os.path.normpath( path )
        return path

    def decodeArgs( self, args ) :
        # Scan arguments.

        # Print help?
        help = None
        #if len( args ) < 1 : help = 1
        if args.count( '-?' ) :
            args.remove( '-?' )
            help = 1
        elif args.count( '-help' ) :
            args.remove( '-help' )
            help = 1
        elif args.count( '/help' ) :
            args.remove( '/help' )
            help = 1
        elif args.count( '/?' ) :
            args.remove( '/?' )
            help = 1
        if help :
            self.printHelp()

        # Print revision history?
        history = 0
        if args.count( '-history' ) :
            history = 1
            args.remove( '-history' )
        elif args.count( '-ver' ) :
            history = 1
            args.remove( '-ver' )
        elif args.count( '-version' ) :
            history = 1
            args.remove( '-version' )
        if history :
            self.printHistory()

        # EXit if asked for help, revision, or both.
        if help or history : self.exit( 1 )

        # Generate empty LBX commentary?
        try :
            idx = args.index( '-lbxc' )
            del args[ idx ]
            self.m_lbxc = args[ idx ]
            del args[ idx ]
            self.log( 'Building commentary file: %s' % self.m_lbxc )
            return
        except ValueError : pass

        # Generate empty LBX commentary without verses?
        try :
            idx = args.index( '-lbxcnov' )
            del args[ idx ]
            self.m_lbxcnov = args[ idx ]
            del args[ idx ]
            self.log( 'Building commentary file without verses: %s' % 
                self.m_lbxcnov )
            return
        except ValueError : pass

        # LBX and ST tagging (must preceed -st tagging)?
        try :
            idx = args.remove( '-lbxst' )
            self.m_lbxst = 1
            self.m_st = 1
            self.log( 'Libronix and SpiritAndTruth.org web Bible tagging.' )
            if not self.m_st : self.m_tag = ''
        except ValueError : pass

        # LBX and ST tagging (must preceed -st tagging)?
        try :
            idx = args.remove( '-lbxst+' )
            self.m_lbxstplus = 1
            self.m_st = 1
            self.log( 'Libronix and SpiritAndTruth.org web Bible tagging.' )
            if not self.m_st : self.m_tag = ''
        except ValueError : pass

        # SpiritAndTruth abbreviation flag.
        # Must be processed first.
        try :
            idx = args.remove( '-st' )
            self.m_st = 1
            #if not args.count( '-backup' ) :
                #args.append( '-backup' )
            if not args.count( '-scrub' ) :
                args.append( '-scrub' )
            if not args.count( '-tagjs' ) :
                args.append( '-tagjs' )
            if not args.count( '-plainlinks' ) :
                args.append( '-plainlinks' )
        except ValueError : pass

        # Yes as an override?
        try :
            args.remove( '-yes' )
            self.m_ask = 0
        except ValueError : pass

        # Backup directory?
        try :
            idx = args.index( '-backup' )
            del args[ idx ]
            # If directory name potentially specified.
            if idx < len( args ) :
                next = args[ idx ]
                # If not a flag.
                if '-' != next[ 0 ] :
                    del args[ idx ]
                    self.m_backupDir = next
            if not self.m_backupDir :
                key = 'TEMP'
                if not os.environ.has_key( key ) :
                    key = 'TMP'
                    if not os.environ.has_key( key ) :
                        key = 'TMPDIR'
                if os.environ.has_key( key ) :
                    self.m_backupDir = os.environ[ key ]
            if self.m_backupDir :
                self.m_backupDir = self.path( self.m_backupDir )
                if ( os.path.isdir( self.m_backupDir ) ) :
                    self.log( 'Backing up to: %s' % self.m_backupDir )
                else :
                    self.log( 'Backup not a directory: %s' % self.m_backupDir )
                    self.m_backupDir = ''
                    self.exit( 1 )
        except ValueError : pass

        # Assume Book name supplied?
        try :
            idx = args.index( '-book' )
            del args[ idx ]
            self.m_book = args[ idx ]
            del args[ idx ]
            self.log( 'Assumed book: %s' % self.m_book )
        except ValueError : pass

        # Book name file supplied?
        try :
            idx = args.index( '-books' )
            del args[ idx ]
            self.m_booksFileName = args[ idx ]
            del args[ idx ]
            file = None 
            path = None
            try :
                path = self.path( self.m_booksFileName )
                file = open( path, 'r' )
                self.m_books = []
                for line in file.readlines() :
                    line = re.sub( '\s', '', line )
                    self.m_books.append( line )
                if 66 != len( self.m_books ) :
                    self.log( 'Bookfile: % must contain 66 lines (1 line per bible book)!' % self.m_booksFileName )
                    self.exit( 1 )
                self.log( 'Using book file: %s %s' % self.m_booksFileName, self.m_books )
            finally :
               if file : file.close()
        except IOError :
            self.log( 'Unable to read: %s (%s)' % ( self.m_booksFileName, path ) )
            self.exit( 1 )
        except ValueError : pass


        # SwordSearcher tagging?
        try :
            idx = args.remove( '-swordsearcher' )
            self.m_swordsearcher = 1
            self.m_range = 1
            self.m_tagbook = 1
            self.log( 'SwordSearcher tagging.' )
            self.m_tag = [ '<SCRIPREF>', '</SCRIPREF>' ]
        except ValueError : pass

        # Olivetree tagging?
        try :
            idx = args.remove( '-olivetree' )
            self.m_olivetree = 1
            self.log( 'Olivetree tagging.' )
            self.m_tag = [ '<xb=', 'b.c.v>', '</xb>' ]
        except ValueError : pass

        # Libronix tagging?
        try :
            idx = args.remove( '-lbx' )
            self.m_lbx = 1
            self.log( 'Libronix tagging.' )
            self.m_tag = ''
        except ValueError : pass

        # Libronix PBB tagging?
        try :
            idx = args.remove( '-lbxpbb' )
            self.m_lbxpbb = 1
            self.log( 'Libronix PBB tagging.' )
            self.m_tag = ''
        except ValueError : pass

        # LBX and Gospelcom tagging?
        try :
            idx = args.remove( '-lbxgospel' )
            self.m_lbxgospel = 1
            self.log( 'Libronix and bible.gospelcom.net web Bible tagging.' )
            self.m_tag = ''
        except ValueError : pass

        # LBX and Gospelcom tagging?
        try :
            idx = args.remove( '-lbxgospel+' )
            self.m_lbxgospelplus = 1
            self.log( 'Libronix and bible.gospelcom.net web Bible tagging.' )
            self.m_tag = ''
        except ValueError : pass

        # LBX and Studylight tagging?
        try :
            idx = args.remove( '-lbxstudy' )
            self.m_lbxstudy = 1
            self.log( 'Libronix and www.studylight.org web Bible tagging.' )
            self.m_tag = ''
        except ValueError : pass

        # LBX and Studylight tagging?
        try :
            idx = args.remove( '-lbxstudy+' )
            self.m_lbxstudyplus = 1
            self.log( 'Libronix and www.studylight.org web Bible tagging.' )
            self.m_tag = ''
        except ValueError : pass

        # Gospelcom tagging?
        try :
            idx = args.remove( '-gospelcom' )
            self.m_gospelcom = 1
            self.log( 'Gospelcom tagging.' )
            self.m_tag = ''
        except ValueError : pass

        # Studylight tagging?
        try :
            idx = args.remove( '-studylight' )
            self.m_studylight = 1
            self.log( 'Studylight tagging.' )
            self.m_tag = ''
        except ValueError : pass

        # Exact matching?
        try :
            idx = args.remove( '-exact' )
            self.m_exactMatch = address.AddressFinder.MATCH_EXACT
            self.log( 'Exact pattern matching.' )
        except ValueError : pass

        # Exact matching?
        try :
            idx = args.remove( '-ignorecase' )
            self.m_exactMatch = address.AddressFinder.MATCH_IGNORECASE
            self.log( 'Ignoring case in pattern matching.' )
        except ValueError : pass

        # Insert file supplied?
        try :
            idx = args.index( '-insert' )
            del args[ idx ]
            self.m_insertFileName = args[ idx ]
            del args[ idx ]
            file = None 
            path = None
            try :
                path = self.path( self.m_insertFileName )
                file = open( path, 'r' )
                self.m_books = []
                self.m_insertFileContents = file.readlines()
                log ( 'Inserting file:', self.m_insertFileName )
            finally :
               if file : file.close()
        except IOError :
            self.log( 'Unable to read: %s (%s)' % ( self.m_insertFileName, path ) )
            self.exit( 1 )
        except ValueError : pass


        # Verbose?
        try :
            idx = args.remove( '-v' )
            self.m_verbose = 1
            self.log( 'Verbose.' )
        except ValueError : pass

        # Preserve joined lines in output?
        try :
            idx = args.remove( '-join' )
            self.m_join = 1
            self.log( 'Preserve joined lines in output.' )
        except ValueError : pass

        # Don't Join lines?
        try :
            idx = args.remove( '-nojoin' )
            self.m_nojoin = 1
            self.log( 'Do not join lines before scanning addresses.' )
        except ValueError : pass

        # Scrub lines?
        try :
            idx = args.remove( '-scrub' )
            self.m_scrub = 1
            self.log( 'Scrub lines before tagging.' )
        except ValueError : pass

        # Chapters lacking verse use entire verse range?
        try :
            idx = args.remove( '-range' )
            self.m_range = 1
            self.log( 'Chapters lacking verse use entire verse range.' )
        except ValueError : pass

        # Custom file extension?
        try :
            idx = args.index( '-ext' )
            del args[ idx ]
            self.m_ext = args[ idx ]
            del args[ idx ]
        except ValueError: pass
        self.log( 'Scan for file extension: %s' % self.m_ext )

        # Javascript tag?
        try :
            idx = args.index( '-tagjs' )
            self.m_tagjsPath = self.m_defaultLookup
            del args[ idx ]
            if idx < len( args ) :
                next = args[ idx ]
                # If not a flag.
                if '-' != next[ 0 ] :
                    del args[ idx ]
                    self.m_tagjsPath = next
            self.log( 'Mark with javascript tag using path: %s' % self.m_tagjsPath )
            self.m_tag = [ '<a href=\"javascript:X(\'', '\')">', '</a>' ]
        except ValueError : pass

        # Custom tag?
        if not self.m_tagjsPath :
            try :
                idx = args.index( '-tag' )
                del args[ idx ]
                tag = args[ idx ]
                parts = tag.split( '|' )
                if 3 != len( parts ) :
                    self.log( 'Invalid tag - must contain 3 fields separated by | ' )
                    self.log(  tag )
                    self.exit( 1 )
                else :
                    self.m_tag = tag
                    self.log( 'Tag: %s' % string.join( self.m_tag,  '|' ) )
                del args[ idx ]
            except ValueError : pass


        # Custom variable tag?
        try :
            idx = args.index( '-tagvar' )
            del args[ idx ]
            self.m_tagvar = args[ idx ]
            del args[ idx ]
            self.log( 'Tag: %s' % self.m_tagvar )
            # Are we going to include bible tagging?
            addrVar = re.search( '%%(QUOTED)?(BOOK|BOOKLONG|CHAPTER|VERSE|BIBLEADDRESS)%%', self.m_tagvar )
            if addrVar : self.m_tagvaraddr = self.m_tagvar
        except ValueError : pass

        # Plain links for URLs?
        try :
            idx = args.remove( '-plainlinks' )
            self.m_plainlinks = '''
            <style>
            <!-- 
            a:link { text-decoration: none; }
            a:visited {  text-decoration: none; }
            a:hover { text-decoration: underline; }
            -->
            </style>
            '''
            self.log( 'Plain links (not underlined).' )
        except ValueError : pass

        # Include book names in tagging?
        try :
            idx = args.remove( '-tagbook' )
            self.m_tagbook = 1
            self.log( 'Tag surrounds book name.' )
        except ValueError : pass


        # Topic tags supplied?
        filename = ''
        file = None
        try :
            idx = args.index( '-topics' )
            del args[ idx ]
            try :
                filename = args[ idx ]
                del args[ idx ]
                self.log( 'Topic file: %s' % filename )
                path = self.path( filename )
                file = open( path, 'r' )
                for line in file.readlines() :
                    line = re.sub( '\s+$', '', line )
                    self.log( 'Topic: %s' % line )
                    line = re.sub( '"', '\\"', line )
                    line = re.sub( "'", "\\'", line )
                    line = re.sub( '\(', '\\(', line )
                    line = re.sub( '\)', '\\)', line )
                    #line = re.sub( '.', '\.', line )
                    line = re.sub( '\*', '\\S*', line )
                    line = re.sub( r'\?', '.', line )
                    line = re.sub( r'\\', '\\\\', line )
                    self.m_topics.append( line )

                # If no variable tagging specification provided.
                if not self.m_tagvar :

                    #lbxTopicLink = '<a title="Lookup %%TOPIC%% in Logos Bible Software Series X" href="libronixdls:keylink|ref=topic%2Btopics.%%QUOTEDTOPIC%%|res=LLS:14.0.24">%%MATCH%%</a>'
                    #lbxTopicLink = '<a title="Lookup %%TOPIC%% in Logos Bible Software Series X" href="libronixdls:jump|ref=topic%2Btopics.%%QUOTEDTOPIC%%|res=LLS:14.0.24">%%MATCH%%</a>'
                    # User will need to create a keylink from "Text" data type to their favorite dictionaries.
                    lbxTopicLink = '<a title="Lookup %%TOPIC%% in Logos Bible Software Series X" href="libronixdls:keylink|ref=[en]text:%%QUOTEDTOPIC%%">%%MATCH%%</a>'

                    #lbxTopicLinkIcon = '<a title="Lookup %%TOPIC%% in Logos Bible Software Series X" href="libronixdls%3:keylink|ref=topic%2Btopics.%%QUOTEDTOPIC%%|res=LLS:14.0.24"><img border="0" src="LibronixLink.gif"></a>'
                    #lbxTopicLinkIcon = '<a title="Lookup %%TOPIC%% in Logos Bible Software Series X" href="libronixdls:jump|ref=topic%2Btopics.%%QUOTEDTOPIC%%|res=LLS:14.0.24"><img border="0" src="LibronixLink.gif"></a>'
                    lbxTopicLinkIcon = '<a title="Lookup %%TOPIC%% in Logos Bible Software Series X" href="libronixdls:keylink|ref=[en]text:%%QUOTEDTOPIC%%"><img border="0" src="LibronixLink.gif"></a>'
                    lbxTopicLinkPlus = '<a title="Lookup %%TOPIC%% in Logos Bible Software Series X" href="libronixdls:jump|ref=[en]text:%%QUOTEDTOPIC%%|res=LLS:14.0.24">+</a>'

                    gospelTopicLink = '<a title="Lookup %%TOPIC%% at Gospelnet.com" href="https://bible.gospelcom.net/cgi-bin/bible?SearchType=AND&language=english&searchpage=0&search=%%QUOTEDTOPIC%%">%%MATCH%%</a>'
                    studyTopicLink = '<a title="Lookup %%TOPIC%% at StudyLight.org" href="https://www.studylight.org/desk/?l=en&query=%%QUOTEDTOPIC%%">%%MATCH%%</a>'
                    stTopicLink = '<a title="Lookup %%TOPIC%% at SpiritAndTruth.org" href="https://www.SpiritAndTruth.org/lookup/index.htm?%%QUOTEDTOPIC%%">%%MATCH%%</a>'

                    if self.m_lbx :
                        self.m_tagvar = lbxTopicLink
                    elif self.m_lbxgospel :
                        self.m_tagvar = gospelTopicLink
                        self.m_tagvar += lbxTopicLinkIcon
                    elif self.m_lbxgospelplus :
                        self.m_tagvar = gospelTopicLink
                        self.m_tagvar += lbxTopicLinkPlus
                    elif self.m_lbxstudy :
                        self.m_tagvar = studyTopicLink
                        self.m_tagvar += lbxTopicLinkIcon
                    elif self.m_lbxstudyplus :
                        self.m_tagvar = studyTopicLink
                        self.m_tagvar += lbxTopicLinkPlus
                    elif self.m_lbxpbb :
                        #self.m_tagvar = '[[%%MATCH%%|Topic:%%TOPIC%%]]'
                        self.m_tagvar = '[[%%MATCH%%|topic.%%TOPIC%%]]'
                    elif self.m_lbxst :
                        self.m_tagvar = stTopicLink
                        self.m_tagvar += lbxTopicLinkIcon
                    elif self.m_lbxstplus :
                        self.m_tagvar = stTopicLink
                        self.m_tagvar += lbxTopicLinkPlus
                    elif self.m_st :
                        self.m_tagvar = stTopicLink
                file.close()
            except valueError :
                self.log( 'Missing topic file after -topics flag!' )
                self.exit( 1 )
        except ValueError : pass

        return args

    def exit( self, value ) :
        sys.exit( value )

    def printHelp( self ) :
        # Print help instructiions.

            global VERSION
            print 'bibletag ver.', VERSION
            print 'address ver.', address.VERSION
            global HELP
            print HELP
            return

    def printHistory( self ) :
        # Print revision history.

            global HISTORY
            print 'bibletag ver.', VERSION
            print 'address ver.', address.VERSION
            print HISTORY
            print '----------------------------------------------------'
            print '    ADDRESS SCANNING HISTORY'
            print '----------------------------------------------------'
            print address.HISTORY
            return

    def scanFile( self, path ) :
        # Scan file and tag for bible lookup.

        # Reject files without the specified extension.
        try :
            pathExt = os.path.splitext( path )[ 1 ]
            filterExt = re.compile( self.m_ext + '$', re.IGNORECASE )
            if not filterExt.match( pathExt ) : return

        # No extension found.
        except ValueError : return

        # Optionally ask before tagging each file.
        if ( self.m_ask ) :
            answer = raw_input( 'Tag file: ' + path + '? [y] ' )
            if answer and 'y' != answer[ 0 ] and 'Y' != answer[ 0 ] : 
                self.log( 'Skipping: %s' % path )
                return

        # Compute relative path from the specified directory.
        walkDir = os.path.dirname( self.m_walkFromPath )
        sharedPath = os.path.commonprefix( [ walkDir, path ] )
        relativePath = path[ len( sharedPath ) + 1 : ]

        # Make backup?
        if ( self.m_backupDir ) :
            backupFile = os.path.join( self.m_backupDir, relativePath )
            backupDir = os.path.dirname( backupFile )
            # Create directory if not already there.
            try :
                os.makedirs( backupDir )
            # Directory may already exist.
            except OSError : pass

            self.log( 'Backing up to: %s' % backupFile )
            shutil.copy2( path, backupFile )

        # Modify relative URL in tag?
        if self.m_tag : self.m_tag[ 0 ] = self.relativeTagUrl( self.m_tag[ 0 ] )
        if self.m_tagvar : self.m_tagvar = self.relativeTagUrl( self.m_tagvar )
        if self.m_tagvaraddr : self.m_tagvaraddr = self.relativeTagUrl( self.m_tagvaraddr )

        tagFile = None
        try :

            # Open the tag file.
            tagFile = open( path, 'r' )

            # Does the file require a <body marker?
            ext = self.m_ext[ 0 : 4 ].lower()
            bodyFound = 0
            bodyRequired = ( '.htm' == ext or '.xml' == ext )

            # Prepare to build output.
            outLines = []
            # For each line in the file.
            lineNumber = 0
            startJoinNumber = 1
            body = 0
            joined = ''
            inserted = 0
            lines = tagFile.readlines()
            tagFile.close()
            for line in lines :
                lineNumber = lineNumber + 1

                # If this file should contain a <body marker.
                if bodyRequired:

                    # If found "<body. . ."
                    match = self.m_insertExp.search( line )
                    if match :
                        bodyFound = 1

                        # Pass through up to '<body...'
                        expanded = line[ 0 : match.start() ]
                        expanded = expanded + '\n'

                        # Insert optional text?
                        if not inserted :
                            inserted = 1

                            # Insert javscript tag support code?
                            if self.m_tagjsPath :
                                depth = len( relativePath.split( os.sep ) ) - 2
                                urlPath = self.m_tagjsPath
                                # If path is relative.
                                if not re.search( ':', urlPath ) :

                                    # Adjust relative path.
                                    while ( 0 < depth ) : 
                                        urlPath =  '../' + urlPath
                                        depth = depth - 1

                                urlDir= os.path.dirname( urlPath )
                                self.log( 'Inserting -tagjs code using path: %s' % urlPath )
                                expanded = expanded + self.m_tagjsCode

                            # Insert user-specified file?
                            if self.m_insertFileContents :
                                self.log( 'Inserting user-specified file: %s' % self.m_insertFileName )

                                # Insert our file contents
                                expanded = expanded + string.join( self.m_insertFileContents, '' )

                            # Expand variables
                            if self.m_tagjsPath or self.m_insertFileContents :
                                expanded = re.sub( '%URLPATH%', urlPath, expanded )
                                expanded = re.sub( '%URLDIR%', urlDir, expanded )

                        # Pass through '<body...'
                        line = expanded + line[ match.start() : ]

                    # else if body not yet found in html or xml file
                    elif not bodyFound :

                        # Force line to be passed without joining.
                        joined = joined + line
                        line = ''
            
                # Keep track of joins.
                offset = 0
                joins = []

                # If not joining or last line or blank line, flush pending joins. 
                if self.m_nojoin or ( line is lines[ -1 ] or re.match( '^\s*$', line ) ) :
                    line = joined + line
                    joined = ''

                # Else a non blank line prior to the end.
                else :
                
                    # Join line with others.
                    joined = joined + line
                    offset = offset + len( line )
                    joins.append( offset )
                    continue

                # Scrub lines if needed.
                for scrub in self.m_scrubs :
                    prev = line
                    expr, replace = scrub
                    line = expr.sub( replace, line )
                    #if prev != line :
                    #    assert 0

                # Suppress link underlining?
                if self.m_plainlinks :
                    old = line
                    #line = re.sub( '(<head>(?i))', r'\1\n' + self.m_plainlinks, line )
                    new = re.sub( '(</head>(?i))', self.m_plainlinks + r'\n\1\n', line )
                    if old == new :
                        #line = re.sub( '(<html>(?i))', r'\1\n<head>\n' + self.m_plainlinks + r'\n</head>\n', line )
                        new = re.sub( '(<body>(?i))', r'\1<head>\n' + self.m_plainlinks + r'\n</head>\n', line )
                    if old != new :
                        self.m_plainlinks = ''
                    line = new

                # If a body is required and we haven't seen it yet.
                if bodyRequired and not bodyFound :

                    # Pass line to output without tagging.
                    outLines.append( line )

                # Else no body required or we've seen it.
                else :

                    # Tag all the addresses.
                    tagged = self.tagLine( '%d-%d' % ( startJoinNumber, lineNumber ), line )
                    startJoinNumber = lineNumber

                    # If not preserving joined lines, separate lines.
                    lastjoin = 0
                    if not self.m_join : 
                        separated = ''
                        for join in joins:
                            separated = separated + tagged[ lastjoin : join ]
                            separated = separated + '\n'
                            lastjoin = join
                        separated = separated + tagged[ lastjoin : ]

                    # Output the result.
                    outLines.append( separated )


            # Default operation creates new parallel file.
            base, ext = os.path.splitext( path )
            outpath = base + '_bibletag' + ext

            # If a backup directory was specified.
            if self.m_backupDir :

                # Now that we've completed tagging successfully,
                # save the output back over the input file.
                outpath = path

            # Save output to parallel file.
            tagFile = open( outpath, 'w' )
            self.log( 'Writing: %s' % outpath )

            #os.lseek( tagFile.fileno(), 0, 0 )
            tagFile.writelines( outLines )

            if self.m_insertFileContents and not inserted :
                self.log( 'Line %s: WARNING: no <body> found, insertion not made!' % lineNumber )

        finally :
            if tagFile : tagFile.close()

    def relativeTagUrl( self, tagUrl ) :
        useUrl = tagUrl
        match = self.m_relUrlExp.search( tagUrl )
        if match:
            depth = len( relativePath.split( os.sep ) ) - 2
            prefix = ''
            while ( depth ) : 
                prefix = prefix + '../'
                depth = depth - 1
            #prefix = '<prefix>'
            if prefix :
                #useUrl = self.m_urlIntroExp.sub( r'\1\2' + prefix, tagUrl )
                useUrl = self.m_urlIntroExp.sub( r'\1' + prefix, tagUrl )
                self.log( 'Changing relative url:\nwas: %s\nnow: %s' % ( tagUrl, useUrl ) )
        #assert 0
        return useUrl

    def htmlPurge( self, line ) :
        # Purge HTML from line.
        start = 0

        # Convert mdash and ndash to dash prior to address matching.
        old = line
        purged = re.sub( self.m_mndash, '-', line  )

        # Convert from immutable string to list so we can
        # replace HTML with spaces in-place.
        purged = list( purged )

        # Get rid of html keywords
        while 1 :
            match = self.m_htmlExp.search( line, start )
            if match :
                # Keep the line the same length by replacing
                # html by equivalent number of spaces.
                # Otherwise, location of addresses found in this
                # purged string will not be accurate.
                for i in range( match.start(), match.end() ) : purged[ i ] = ' '
                start = match.end()
            else :
                break
        purged = string.join( purged, '' )
        return purged

    def lbxPurge( self, line ) :
        # Purge LBX from line.
        start = 0

        # Convert mdash and ndash to dash prior to address matching.
        old = line

        # Convert from immutable string to list so we can
        # replace LBX milestones and links with another character in-place.
        purged = list( line )

        # Get rid of lbx tags (we erase them so they don't get retagged)
        while 1 :
            match = self.m_lbxpbbExp.search( line, start )
            if match :
                # Keep the line the same length by replacing
                # html by equivalent number of spaces.
                # Otherwise, location of addresses found in this
                # purged string will not be accurate.
                for i in range( match.start(), match.end() ) : purged[ i ] = '#'
                start = match.end()
            else :
                break
        purged = string.join( purged, '' )
        return purged

    # No longer used because we throw away the purged info
    # after we've located the addresses so no need to fold
    # HTML back in.
    def htmlMerge_RETIRED( self, line, htmls ) :
        # Merge HTML back into purged lines.
        merged = line
        for html in htmls :
            start, text = html
            end = start + len( text )
            merged = merged[ : start ] + text + merged[ end : ]
        return merged

    def purgeMarkup( self, line ) :

        # Purge all Libronix milestones and tags?
        purged = line
        if self.m_lbxpbb : purged = self.lbxPurge( line )

        # Purging should not change length.
        assert len( purged ) == len( line )

        # Purge all HTML except anchors.
        purged = self.htmlPurge( purged )

        # Purging should not change length.
        assert len( purged ) == len( line )

        return purged

    def tagLine( self, lineNumber, line ) :
        # Tag all the addresses in a line.

        # Purge all markup.
        purged = self.purgeMarkup( line )

        # Tag Topics?
        if self.m_topics :

            # Make sure we have a proper topic tag string.
            if not self.m_tagvar or not re.search( '%%TOPIC%%', self.m_tagvar ) :
                self.log( '-topics flag requires additional -tagvar string containing %%TOPIC%% variable' )
                assert 0

            # For each topic entry.
            for topicLine in self.m_topics :

                # For each topic phrase.
                phrases = topicLine.split( '|' )
                topic = phrases[ 0 ]
                topicMarker = self.protectTopic( topic )
                topicMarkerQuoted = self.protectTopic( urllib.quote( topic ) )

                # Match longest first.
                sorted = []
                for phrase in phrases :
                    sorted.append( ( len( phrase ), phrase ) )
                sorted.sort()
                sorted.reverse()
                phrases = []
                for sort in sorted :
                    length, phrase = sort
                    phrases.append( phrase )

                for phrase in phrases :
                    phrase = re.sub( ' ', r'\s+', phrase )
                    lOut = pOut = ''
                    iStart = 0
                    phrase = r'\b%s\b' % phrase
                    topicMatchExp = re.compile( phrase, re.IGNORECASE )

                    # Find and tag all occurrences of the phrase.
                    while( 1 ) :
                        tMatch = topicMatchExp.search( purged, iStart )
                        if tMatch :
                            matchMarker = self.protectTopic( tMatch.group() )
                            matchMarkerQuoted = self.protectTopic( urllib.quote( tMatch.group() ) )

                            # Let user know?
                            if self.m_verbose : 
                                self.log( 'Line: %s, topic: "%s" tagged as "%s"' % ( lineNumber, tMatch.group(), topic ) )

                            # Pass through part up to the match.
                            lOut += line[ iStart : tMatch.start() ]
                            pOut += purged[ iStart : tMatch.start() ]

                            # Place the tag.
                            tag = re.sub( '%%TOPIC%%', topicMarker, self.m_tagvar )
                            tag = re.sub( '%%QUOTEDTOPIC%%', topicMarkerQuoted, tag )
                            tag = re.sub( '%%MATCH%%', matchMarker, tag )
                            tag = re.sub( '%%QUOTEDMATCH%%', matchMarkerQuoted, tag )
                            lOut += tag
                            pOut += tag

                            # Move to next portion.
                            iStart = tMatch.end()

                        else :

                            # Preserve remainder (may be entire line if no Strong's).
                            line = lOut + line[ iStart : ]
                            purged = pOut + purged[ iStart : ]
                            break

            # Delete markers which protect against multiple expansion.
            line = re.sub( self.m_topicMarker, '', line )
            purged = re.sub( self.m_topicMarker, '', purged )

            # Purge all markup (may have added more with our own tags).
            purged = self.purgeMarkup( line )

        # Tag Strong's numbers if LBX format.
        if self.m_lbxpbb :
            lOut = pOut = ''
            iStart = 0
            while 1 :
                sMatch = self.m_lbxpbbStrongsExp.search( purged, iStart )
                if sMatch :

                    # Let user know?
                    if self.m_verbose : 
                        self.log( '\nStrongs: %s' % sMatch.group() )

                    # Pass through part up to the match.
                    lOut += line[ iStart : sMatch.start() ]
                    pOut += purged[ iStart : sMatch.start() ]

                    # Place the tag.
                    language = 'Hebrew'
                    prefix = sMatch.group( 'prefix' )
                    number = sMatch.group( 'number' )
                    if 'g' == prefix.lower() : language = 'Greek'
                    tag = '[[%sStrongs:%s%s]]' % (language, prefix, number)
                    lOut += tag
                    pOut += tag


                    # Move to next portion.
                    iStart = sMatch.end()

                else :

                    # Preserve remainder (may be entire line if no Strong's).
                    line = lOut + line[ iStart : ]
                    purged = pOut + purged[ iStart : ]
                    break

            # Purge all markup (may have added more with our own tags).
            purged = self.purgeMarkup( line )

        # Find addresses in clean source.
        addresses = self.m_addressFinder.findAddresses( lineNumber, purged, self.m_range, self.m_tagbook )

        iStart = 0
        output = ''
        prevLine = line
        for address in addresses :

            # Pass through text before the address.
            addressStart = address.start()
            output = output + line[ iStart : addressStart ]

            # If SwordSearcher tagging.
            if self.m_swordsearcher :

                # Place swordsearcher tag.
                book = re.sub( '\.', '', address.book() )
                output = output + '<SCRIPREF>%s %s:%s</SCRIPREF>' % ( book, address.chapter(), address.verseRange() )

            # Else if olivetree tagging.
            elif self.m_olivetree :

                # Place olivetree tag.
                output = output + '<xb=%d.%s.%s>%s</xb>' % \
                    ( address.bookNumber(), address.chapter(), address.verseRange(), address.original() )

            # Else if gospelcom tagging.
            elif self.m_studylight :
                output = output + address.studylightLink()

            # Else if gospelcom tagging.
            elif self.m_gospelcom :
                output = output + address.gospelcomLink()

            # Else if Libronix tagging.
            elif self.m_lbx :
                output = output + address.lbxLaunchLink()

            # Else if Libronix PBB tagging.
            elif self.m_lbxpbb :
                output = output + address.lbxLink()

            # Else if Libronix and Gospel tagging.
            elif self.m_lbxgospel or self.m_lbxgospelplus:

                img = None
                if self.m_lbxgospelplus : img = '+'
                output = output + address.gospelcomLink()
                output = output + address.lbxLaunchLinkImg( img )

            # Else if Libronix and Gospel tagging.
            elif self.m_lbxstudy or self.m_lbxstudyplus:

                img = None
                if self.m_lbxstudyplus : img = '+'
                output = output + address.studylightLink()
                output = output + address.lbxLaunchLinkImg( img )

            # Else normal (ST) tagging.
            else :

                # If using string tag which includes verse variables.
                if self.m_tagvaraddr :

                    # Expand variables in the tag.
                    tag = self.m_tagvar
                    tag = re.sub( '%%BOOK%%', address.book() , tag )
                    tag = re.sub( '%%BOOKLONG%%', address.book() , tag )
                    tag = re.sub( '%%CHAPTER%%', address.chapter() , tag )
                    tag = re.sub( '%%VERSE%%', address.verseRange() , tag )
                    tag = re.sub( '%%BIBLEADDRESS%%', address.short() , tag )
                    tag = re.sub( '%%MATCH%%', address.original(), tag )

                    tag = re.sub( '%%QUOTEDBOOK%%', urllib.quote( address.book() ) , tag )
                    tag = re.sub( '%%QUOTEDBOOKLONG%%', urllib.quote( address.book() ) , tag )
                    tag = re.sub( '%%QUOTEDCHAPTER%%', urllib.quote( address.chapter() ) , tag )
                    tag = re.sub( '%%QUOTEDVERSE%%', urllib.quote( address.verseRange() ) , tag )
                    tag = re.sub( '%%QUOTEDBIBLEADDRESS%%', urllib.quote( address.verseRange() ) , tag )
                    tag = re.sub( '%%QUOTEDMATCH%%', urllib.quote( address.original() ) , tag )

                    # Place tag.
                    output = output + tag

                # Else using standard 3-part tag.
                else :

                    # Place first part of tag.
                    output = output + self.m_tag[ 0 ]

                    # Place URL portion of tag.
                    #anchor = '%s_C%sV%s' % ( address.book(), address.chapter(), address.verseRange() )
                    anchor = '%s %s:%s' % ( address.book(), address.chapter(), address.verseRange() )
                    output = output + urllib.quote( anchor )

                    # Place middle of tag.
                    output = output + self.m_tag[ 1 ]

                    # Place the original representation of the address.
                    original = address.original()
                    output = output + original

                    # End of tag.
                    output = output + self.m_tag[ 2 ]

                # Else if Libronix and ST tagging.
                if self.m_lbxst or self.m_lbxstplus :
                    img = None
                    if self.m_lbxstplus : img = '+'
                    output = output + address.lbxLaunchLinkImg( img )

            # Move past the address.
            iStart = addressStart + address.length()

        # Preserve remainder (may be entire line if no addresses).
        line = output + line[ iStart : ]

        return line

    def protectTopic( self, topic ) :
        # Protect topic from subsequent match expansion.
        return '%s%s%s' % ( self.m_topicMarker, topic, self.m_topicMarker )

# If module is launched as a program, perform indexing using specified files.
if __name__ == '__main__' :
    tag = Processor( sys.argv )
    
